@@ -15,7 +15,7 @@ itcl_class InputDialog {
 
 	# add an entry widget below the message
 	global $this.choice
-	pack [entry $this.entry] \
+	pack [entry $this.entry -relief sunken] \
 		-in $this.top.ext -side left -fill x -expand 1 \
 		-padx 2m -pady 2m -ipady 1m
 	bind $this.entry <Return> \
