@@ -30,7 +30,7 @@ itcl_class TableList {
     # format and return the table headings
 
     method format_headings_ {h} {
-	return [eval "format {$formats} $h"]
+	return [eval "format {$hformats} $h"]
     }
 
 
@@ -42,7 +42,6 @@ itcl_class TableList {
 	if {$static_col_sizes && [info exists size_(1)]} {
 	    return
 	}
-
 	for {set c 1} {$c <= $num_cols_} {incr c} {
 	    set size_($c) $hsize_($c)
 	}
@@ -77,6 +76,9 @@ itcl_class TableList {
     # The formats list uses %n$-s type format strings to set left/right 
     # alignment, width and order all at once, for example:
     # "%-10s" for left justify, "%.0s" effectively hides the item...
+    #
+    # If the Precision option is set for a column, assume it is a floating
+    # point value to be formatted like: %6.2f for example (precision = 2)
 
     method calculate_format_ {} {
 	if {$formats_flag_} {
@@ -85,6 +87,7 @@ itcl_class TableList {
 
 	get_col_sizes_
 
+	set hformats {}
 	set formats {}
 	set order_ {}
 	foreach name [$this.config_file info public order -value] {
@@ -94,6 +97,7 @@ itcl_class TableList {
 	    set align [$this.config_file get_option $name Align]
 	    set show [$this.config_file get_option $name Show]
 	    set sep [$this.config_file get_option $name Separator]
+	    set prec [$this.config_file get_option $name Precision]
 
 	    if {"$align" == "Left"} {
 		set a -
@@ -101,10 +105,20 @@ itcl_class TableList {
 		set a {}
 	    }
 
-	    if {!$show} {
-		append formats [format {%%%d$.0s} $col]
+	    if {$prec == 0} {
+		set p {}
+		set t s
 	    } else {
-		append formats [format {%%%d$%s%ds %s } $col $a $size_($col) $sep]
+		set p .$prec
+		set t f
+	    }
+
+	    if {!$show} {
+		append formats [set f [format {%%%d$.0s} $col]]
+		append hformats $f
+	    } else {
+		append formats [format {%%%d$%s%d%s%s %s } $col $a $size_($col) $p $t $sep]
+		append hformats [format {%%%d$%s%ds %s } $col $a $size_($col) $sep]
 	    }
 	}
     }
@@ -151,23 +165,6 @@ itcl_class TableList {
     }
 
 
-    # set up the row range format string for the table header, including the 
-    # number of rows and the first and last index
-
-    method format_table_range_ {} {
-	if {$show_range && $info_rows_ > 0} {
-	    if {$info_rows_ != $total_rows_} {
-		set s " \[$total_rows_\]"
-	    } else {
-		set s ""
-	    }
-	    set range_format_ "(rows %d - %d of $info_rows_$s)"
-	} else {
-	    set range_format_ ""
-	}
-    }
-
-    
     # sort the info list, if requested, based on the
     # sort keys
 
@@ -177,13 +174,14 @@ itcl_class TableList {
 	}
     }
 
-    
-    # compare the 2 given rows (called by lsort above)
+    # compare the 2 given rows, ascii version (called by lsort above)
     # (see lsort man page for desc.)
+    # Note the "1" prepended to the operands: this is in case you have
+    # numbers that look like "0nnn" by are not octal (try [expr 0510 > 0328]).
 
     method cmp_row_ {r1 r2} {
 	foreach i $sort_by {
-	    if {[set v1 [lindex $r1 $i]] > [set v2 [lindex $r2 $i]]} {
+	    if {"[set v1 1[lindex $r1 $i]]" > "1[set v2 [lindex $r2 $i]]"} {
 		return 1
 	    } elseif {$v1 < $v2} {
 		return -1
@@ -192,7 +190,7 @@ itcl_class TableList {
 	return 0
     }
 
-    
+
     # move the selected row down 1 row
     # and make the changes in the info list
 
@@ -211,8 +209,7 @@ itcl_class TableList {
 	    $listbox_ insert $j $s
 	    set disp_info_ [linsert $disp_info_ $j $r] 
 	}
-	$listbox_ select from [expr [lindex $list 0]+1]
-	$listbox_ select to [expr [lindex $rlist 0]+1]
+	select_rows [expr [lindex $list 0]+1] [expr [lindex $rlist 0]+1]
     }
 
     
@@ -233,8 +230,7 @@ itcl_class TableList {
 	    $listbox_ insert $j $s
 	    set disp_info_ [linsert $disp_info_ $j $r] 
 	}
-	$listbox_ select from [expr [lindex $list 0]-1]
-	$listbox_ select to [expr $i-1]
+	select_rows [expr [lindex $list 0]-1] [expr $i-1]
     }
 
 
@@ -245,41 +241,65 @@ itcl_class TableList {
 	set info {}
 	set disp_info_ {}
 	set info_rows_ 0
-	$this.title config -value ""
-    }
+ 	set total_rows_ 0
+   }
 
-    # add the given range of rows from the info list to the 
-    # table (called while scrolling down first time)
-
-    method add_missing_rows_ {from to} {
-	foreach i [lrange $disp_info_ $from $to] {
-	    $listbox_ insert end [format_line_ $i]
-	}
-	set last_row_ $to
-    }
-
-    
     # replace the contents of the given row with the new info
+    # Note: this assumes that no 2 rows are exactly alike.
+    # We can't use the row index here, since sorting and matching 
+    # may mix things up too much.
 
-    method set_row {index r} {
-	set disp_info_ [lreplace $disp_info_ $index $index $r]
-	$listbox_ delete $index
-	$listbox_ insert $index [format_line_ $r]
+    method set_row {oldrow newrow} {
+	set index [lsearch -exact $info $oldrow]
+	set info [lreplace $info $index $index $newrow]
+	save_yview
+	save_selection
+	busy new_info
+	restore_selection
+	restore_yview
     }
 
-    
     # save the current scroll position so it can be restored
     # later. 
     
     method save_yview {} {
-	set saved_yview_ [lindex [$this.vscroll get] 2]
+	global tk_version
+	if {$tk_version >= 4.0} {
+	    set saved_yview_ [lindex [$listbox_ yview] 0]
+	} else {
+	    set saved_yview_ [lindex [$this.vscroll get] 2]
+	}
     }
 
 
     # restore the previously saved scroll position 
     
     method restore_yview {} {
-	$listbox_ yview $saved_yview_
+	global tk_version
+	if {$tk_version >= 4.0} {
+	    $listbox_ yview moveto $saved_yview_
+	} else {
+	    $listbox_ yview $saved_yview_
+	}
+    }
+
+    
+    # save a list of the currently selected rows so they can be restored
+    # later. 
+    
+    method save_selection {} {
+	set saved_selection_ [$listbox_ curselection]
+    }
+
+
+    # restore the previously saved row selection
+    
+    method restore_selection {} {
+	set n [llength $saved_selection_]
+	if {$n} {
+	    select_rows [lindex $saved_selection_ 0] \
+		[lindex $saved_selection_ [incr n -1]]
+	}
     }
 
 
@@ -299,7 +319,7 @@ itcl_class TableList {
 
     method match_regexp_ {row} {
 	for {set i 0} {$i < $info_cols_} {incr i} {
-	    if {![regexp [lindex $match_list_ $i] [lindex $row $i]]} {
+	    if {![regexp -- [lindex $match_list_ $i] [lindex $row $i]]} {
 		return 0
 	    }
 	}
@@ -307,13 +327,28 @@ itcl_class TableList {
     }
 
 
-    # append a row to the table
-    # Note: you need to do a "new_info" when you are done appending
-    # to calculate the format string (doing it every time would be slow)
+    # return true if the given row matches the current regexp (ignore case)
+
+    method match_regexp_nocase_ {row} {
+	for {set i 0} {$i < $info_cols_} {incr i} {
+	    if {![regexp -nocase -- [lindex $match_list_ $i] [lindex $row $i]]} {
+		return 0
+	    }
+	}
+	return 1
+    }
+
+
+    # append a row to the table. 
+    # Note: you will need to do a "new_info" when you are done appending
+    # rows if you want any further matching, sorting or formatting to occur.
 
     method append_row {row} {
+	$listbox_ insert end [format_line_ $row]
 	lappend info $row
+	lappend disp_info_ $row
 	incr info_rows_
+	incr total_rows_
     }
 
 
@@ -323,13 +358,13 @@ itcl_class TableList {
 	foreach i $rows {
 	    virtual append_row $i
 	}
-	virtual new_info
+	new_info
     }
 
     
     # this method is called whenever the info list changes
 
-    method new_info {} {
+    method new_info {} { 
 	virtual match_info_
 	virtual sort_info_
 	set_info_
@@ -341,14 +376,12 @@ itcl_class TableList {
     method set_info_ {} {
 	calculate_format_
 	set_headings_
-	format_table_range_
 	$listbox_ delete 0 end
-	if {$dynamic_scrolling} {
-	    set n [expr $max_visible_rows_*$scroll_factor]
-	} else {
-	    set n $info_rows_
+
+	set n 0
+	foreach i $disp_info_ {
+	    $listbox_ insert end [format_line_ $i]
 	}
-	virtual add_missing_rows_ 0 $n
     }
 
     
@@ -391,13 +424,29 @@ itcl_class TableList {
 
 
     # select the row with the given index
-    # (make sure the row is present in the listbox)
 
     method select_row {n} {
-	if {$n > $last_row_} {
-	    virtual add_missing_rows_ [expr $last_row_+1] [expr $n+1]
+	global tk_version
+	if {$tk_version >= 4.0} {
+	    $listbox_ selection clear 0 end
+	    $listbox_ selection set $n
+	} else {
+	    $listbox_ select from $n
+	}
+    }
+
+
+    # select a range of rows
+
+    method select_rows {from to} {
+	global tk_version
+	if {$tk_version >= 4.0} {
+	    $listbox_ selection clear 0 end
+	    $listbox_ selection set $from $to
+	} else {
+	    $listbox_ select from $from
+	    $listbox_ select to $to
 	}
-	$listbox_ select from $n
     }
 
 
@@ -441,7 +490,12 @@ itcl_class TableList {
     # clear the selection
 
     method clear_selection {} {
-	$listbox_ select clear
+	global tk_version
+	if {$tk_version >= 4.0} {
+	    $listbox_ selection clear 0 end
+	} else {
+	    $listbox_ select clear
+	}
     }
 
 
@@ -459,14 +513,13 @@ itcl_class TableList {
 	set info $disp_info_
 	new_info
 	if {$n} {
-	    $listbox_ select from $i
-	    $listbox_ select to $i
+	    select_row $i
 	}
 	return $list
     }
 
 
-    # pop up a window to change the layout of the leaf table
+    # pop up a window to change the layout of the table
 
     method layout_dialog {} {
 	if {"$headings" == ""} {
@@ -474,7 +527,6 @@ itcl_class TableList {
 	}
 	busy {
 	    set w [format {%s.tblcfg} [utilGetTopLevel $this]]
-	    set format_flag 0
 	    utilReUseWidget TableListConfig $w \
 		    -table $this
 	}
@@ -518,7 +570,11 @@ itcl_class TableList {
 	}
 	busy {
 	    set w [format {%s.tblprint} [utilGetTopLevel $this]]
-	    utilReUseWidget TableListPrint $w -table $this -printcmd $printcmd
+	    if {[winfo exists $w]} {
+		wm deiconify $w
+	    } else {
+		TableListPrint $w -table $this -printcmd $printcmd
+	    }
 	}
    }
 
@@ -570,54 +626,56 @@ itcl_class TableList {
     #  Align (Left,Right)  - for left or right justify
     #  Separator           - set the separator string (goes after col)
     #  Wildcard            - only show rows where wildcard matches
+    #  Precision           - number of places after the decimal for floating point values
 
     method set_option {name option value} {
 	$this.config_file set_option $name $option $value
     }
 
     
+    # same as set_option, but works on a list of column heading names
+
+    method set_options {headings option value} {
+	foreach name $headings {
+	    $this.config_file set_option $name $option $value
+	}
+    }
+   
+    # return the option value for the given heading name.
+    # See above for list of Options...
+
+    method get_option {name option} {
+	return [$this.config_file get_option $name $option]
+    }
+
+    
     # scroll to the end of the table and display the last set of rows
 
     method show_last_row {} {
-	if {$info_rows_ > $last_row_} {
-	    virtual add_missing_rows_ [expr $last_row_+1] $info_rows_
+	global tk_version
+	if {$tk_version >= 4.0} {
+	    $listbox_ yview moveto 1
+	} else {
+	    # need to update the scrollbar...
+	    update idletasks
+	    $listbox_ yview [expr $info_rows_ - [lindex [$this.vscroll get] 1]]
 	}
-	$listbox_ yview [expr $info_rows_ - $max_visible_rows_ + 1]
     }
 
 
     # scroll both the heading box and the main listbox syncronously
     # (called for horizontal scrolling in listbox)
 
-    method xview {index} {
-	$listbox_ xview $index
-	$headbox_ xview $index
+    method xview {args} {
+	eval "$listbox_ xview $args"
+	eval "$headbox_ xview $args"
     }
 
 
     # scroll the listbox vertically
 
-    method yview {index} {
-	$listbox_ yview $index
-    }
-
-
-    # Try to improve performance for large tables by only inserting
-    # the rows in the table that are actually visible.
-    # (called for vertical scrolling in listbox and at startup)
-
-    method vscroll_ {nrows max first last} {
-	set max_visible_rows_ $max
-	if {$max > 0 && $nrows < $info_rows_} {
-	    if {[set rest [expr $max-($last-$first+1)]] > 0} {
-		set last [min [expr $nrows+$rest] [expr $info_rows_-1]]
-		virtual add_missing_rows_ $nrows [expr $last+($scroll_factor*$max)]
-	    }
-	}
-	$this.vscroll set $info_rows_ $max $first $last
-	if {$show_range} {
-	    $this.title config -value [format $range_format_ [incr first] [incr last]]
-	}
+    method yview {args} {
+	eval "$listbox_ yview $args"
     }
 
 
@@ -693,13 +751,16 @@ itcl_class TableList {
 
     constructor {config} {
 	FrameWidget::constructor
+	global tk_version
 
 	set heading_font [option get . headingFont Tk]
 	set title_font [option get . titleFont Tk]
 	set title_bg [option get . titleBackground Tk]
 
 	# title and menubutton
-	pack [frame $this.top] -side top -fill x
+	pack [frame $this.top] \
+	    -side top -fill x
+
 	if {"$menubar" == ""} {
 	    set menubar $this.top
 	}
@@ -711,28 +772,31 @@ itcl_class TableList {
 	    pack $menubutton_ -side left -padx 1m -ipadx 1m
 	}
 	make_table_menu
-	pack [LabelEntry $this.title \
+
+	pack [label $this.title \
 		  -text $title \
-		  -relief flat \
-		  -state disabled \
-		  -labelFont $title_font] \
-		-side top -in $this.top
+		  -font $title_font] \
+		-side top -fill x -expand 1 -in $this.top
 
 	if {"$title_bg" != ""} {
 	    $this.title config -bg $title_bg
 	}
-	if {$show_range} {
-	    $this.title config -valueWidth 30
-	} else {
-	    $this.title config -valueWidth 0
-	}
 
 	# heading box
-	set headbox_ [listbox $this.headbox \
-		-relief sunken \
-		-font $heading_font \
-		-setgrid 1 \
-		-geometry 1x$heading_lines]
+
+	if {$tk_version >= 4.0} {
+	    set headbox_ [listbox $this.headbox \
+			      -relief sunken \
+			      -font $heading_font \
+			      -setgrid 1 \
+			      -height $heading_lines]
+	} else {
+	    set headbox_ [listbox $this.headbox \
+			      -relief sunken \
+			      -font $heading_font \
+			      -setgrid 1 \
+			      -geometry 1x$heading_lines]
+	}
 
 	# table
 	set hscroll [scrollbar $this.hscroll \
@@ -743,35 +807,27 @@ itcl_class TableList {
 		-relief sunken \
 		-command "$this.listbox yview"]
 	set listbox_ [listbox $this.listbox \
-		-yscroll "$this vscroll_" \
+		-yscroll "$vscroll set" \
 		-xscroll "$hscroll set" \
-		-relief sunken \
-		-setgrid 1 \
-		-exportselection $exportselection]
-
-	if {$single_select} {
-	    tk_listboxSingleSelect $listbox_
-	    bind $listbox_ <B1-Motion> { }
-	    bind $listbox_ <Shift-B1-Motion> { }
-	}
+		-relief sunken]
 
 	# sync scrolling
 	foreach i "$listbox_ $headbox_" {
 	    bind $i <2> "$listbox_ scan mark %x %y; $headbox_ scan mark %x 0"
 	    bind $i <B2-Motion> "$listbox_ scan dragto %x %y; $headbox_ scan dragto %x 0"
 	}
-
 	bind $headbox_ <B1-Motion> { }
 	bind $headbox_ <Shift-B1-Motion> { }
-
 	bind $headbox_ <Shift-1> { }
 	bind $headbox_ <1> { }
-	#bind $headbox_ <Shift-1> "$this sort_dialog"
-	#bind $headbox_ <1> "$this layout_dialog"
 
-	pack $vscroll -side right -fill y
+	if {$show_vscroll} {
+	    pack $vscroll -side right -fill y
+	}
 	pack $headbox_ -side top -fill x
-	pack $hscroll -side bottom -fill x
+	if {$show_hscroll} {
+	    pack $hscroll -side bottom -fill x
+	}
 	pack $listbox_ -fill both -expand 1
 
 	# create an object for reading the config file
@@ -805,21 +861,15 @@ itcl_class TableList {
 	}
     }
 
-    # if true, the range of rows being viewed is displayed in the title
-    public show_range 0 {
-	if {$initialized_} {
-	    if {$show_range} {
-		$this.title.entry config -width 30
-	    } else {
-		$this.title.entry config -width 0
-	    }
-	}
-    }
-
-
     # if true, add a "Table" menu button to the table for config and sorting
     public show_menubutton 1
 
+    # if true, display the vertical scrollbar
+    public show_vscroll 1
+
+    # if true, display the horizontal scrollbar
+    public show_hscroll 1
+
     # name of menubar frame in which to place table menubutton (optional)
     public menubar {}
 
@@ -835,7 +885,9 @@ itcl_class TableList {
 		    if {$heading_lines == 1} {
 			set hsize_($num_cols_) [string length $i]
 		    } else {
-			set hsize_($num_cols_) [max [string length [lindex $i 0]] [string length [lindex $i 1]]]
+			set hsize_($num_cols_) \
+			    [max [string length [lindex $i 0]] \
+				 [string length [lindex $i 1]]]
 		    }
 		    set ord_($i) $num_cols_
 		}
@@ -848,8 +900,11 @@ itcl_class TableList {
     # for 2: each column heading may be a list of 2 (top and bottom) strings.
     public heading_lines 1 {
 	if {$initialized_} {
-	    if {"[utilGetConfigValue $headbox_ -geometry]" != "1x$heading_lines"} {
-		$headbox_ config -geometry 1x$heading_lines
+	    global tk_version
+	    if {$tk_version >= 4.0} {
+		$headbox_ config -height $heading_lines
+	    } else {
+		$headbox_ config -setgrid 1 -geometry 1x$heading_lines
 	    }
 	}
     }
@@ -857,8 +912,13 @@ itcl_class TableList {
     # set the number of rows to be displayed
     public row_lines {} {
 	if {$initialized_} {
-	    if {[scan [utilGetConfigValue $listbox_ -geometry] {%dx%d} w h] == 2} {
-		$listbox_ config -geometry ${w}x$row_lines
+	    global tk_version
+	    if {$tk_version >= 4.0} {
+		$listbox_ config -height $row_lines
+	    } else {
+		if {[scan [utilGetConfigValue $listbox_ -geometry] {%dx%d} w h] == 2} {
+		    $listbox_ config -geometry ${w}x$row_lines
+		}
 	    }
 	}
     }
@@ -872,24 +932,24 @@ itcl_class TableList {
 
     # list of column sizes
     # (if not specified, will be calculated)
-    public sizes {} {
-	set format_flag 0
-    }
+    public sizes {}
 
     # if true, reuse the calculated column sizes rather than recalculate
     # for new info
     public static_col_sizes 0
 
-    # if true, use the scrollbar handler to insert rows dynamically as needed,
-    # otherwise insert all the rows right away
-    public dynamic_scrolling 1
-
     # list of printf formats for columns
-    # (if not specified, will be calculated)
+    # (if not specified, will be calculated, see also -sizes)
     public formats {} {
 	set formats_flag_ 1
+	set hformats $formats
     }
 
+    # print format string for the headings 
+    # (set after -formats, defaults to same as $formats)
+    # This might be different for headings if a column uses %f formats...
+    public hformats {}
+    
     # command to call to filter each row,
     # hook to modify the row before it is displayed
     # arg is is the name of the list holding the row (call by reference)
@@ -898,28 +958,52 @@ itcl_class TableList {
     # default print command
     public printcmd {lpr}
 
-    # flag: single selection mode
-    public single_select {0}
+    # choose the select mode: one of {single, browse, multiple, extended}
+    public selectmode {single} {
+	if {$initialized_} {
+	    global tk_version
+	    if {$tk_version >= 4.0} {
+		$listbox_ config -selectmode $selectmode
+	    } else {
+		config -single_select [expr {"$selectmode" == "single"}]
+	    }
+	}
+    }
+
+    # tk3.6 flag: single selection mode
+    public single_select {0} {
+	if {$initialized_} {
+	    if {$single_select} {
+		tk_listboxSingleSelect $listbox_
+	    }
+	}
+    }
+
 
     # flag: export selection ?
-    public exportselection {1}
+    public exportselection {1} {
+	if {$initialized_} {
+	    $listbox_ config -exportselection $exportselection
+	}
+    }
+
 
     # -- sort options --
 
     # list of col index: sort table based on given columns 
     # (empty means don't sort)
-    public sort_by {}
+    public sort_by {} {
+	if {$initialized_} {
+	    $this.config_file config -sort_by $sort_by
+	}
+    }
 
     # set direction of sort: may be one of (increasing, decreasing) 
-    public sort_order {increasing}
-
-    # set type of sort key: may be one of (ascii, integer, real) 
-    public sort_type {ascii}
-
-    # This number controls how many pages are inserted into the list at a time.
-    # A large number will cause slower startup but make scrolling faster,
-    # while a small number will make startup faster and scrolling slower...
-    public scroll_factor {3} 
+    public sort_order {increasing} {
+	if {$initialized_} {
+	    $this.config_file config -sort_order $sort_order
+	}
+    }
 
     # -- match options --
 
@@ -934,6 +1018,17 @@ itcl_class TableList {
 	}
     }
 
+    # flag: if true, ignore case in matching 
+    # (only works when -use_regexp 1 was specified)
+    public ignore_case 0 {
+	if {$ignore_case} {
+	    set match_any_ ".*"
+	    set match_proc_ match_regexp_nocase_
+	} else {
+	    config -use_regexp $use_regexp
+	}
+    }
+
 
     # -- protected members --
 
@@ -971,9 +1066,6 @@ itcl_class TableList {
     # number of columns in table 
     protected info_cols_ 0
 
-    # last row inserted in the listbox
-    protected last_row_ -1
-
     # length of a line in the table
     protected line_length_ 0
 
@@ -986,12 +1078,12 @@ itcl_class TableList {
     # list of info to display (after filtering and sorting)
     protected disp_info_ {}
 
-    # format string for row range in title, including first and last index and max rows
-    protected range_format_ {}
-
     # used for save_/restore_yview methods
     protected saved_yview_ 0
 
+    # used for save_/restore_selection methods
+    protected saved_selection_ {}
+
     # array(col) of col width
     protected size_
     
@@ -1010,8 +1102,8 @@ itcl_class TableList {
 
 
 # make sure we are using fixed width fontts
-option add Tk*Listbox.font -adobe-courier-medium-r-*-*-*-120-*-*-*-*-*-* widgetDefault
-option add Tk*tableFont -adobe-courier-medium-r-*-*-*-120-*-*-*-*-*-* widgetDefault
-option add Tk*headingFont -adobe-courier-bold-r-*-*-*-120-*-*-*-*-*-* widgetDefault
-option add Tk*titleFont -adobe-courier-bold-r-*-*-*-120-*-*-*-*-*-* widgetDefault
+option add *Listbox.font -adobe-courier-medium-r-*-*-*-120-*-*-*-*-*-* widgetDefault
+option add *tableFont -adobe-courier-medium-r-*-*-*-120-*-*-*-*-*-* widgetDefault
+option add *headingFont -adobe-courier-bold-r-*-*-*-120-*-*-*-*-*-* widgetDefault
+option add *titleFont -adobe-courier-bold-r-*-*-*-120-*-*-*-*-*-* widgetDefault
 
