@@ -14,13 +14,6 @@ proc utilGetTopLevel {w} {
 }
 
 
-# return the value of the given tk option for the given widget
-
-proc utilGetConfigValue {w option} {
-    return [lindex [$w config $option] 4]
-}
-
-
 # get the selection or return an empty string
 
 proc utilGetSelection {} {
@@ -34,22 +27,17 @@ proc utilGetSelection {} {
 # otherwise create it with the given options
 
 proc utilReUseWidget {type w args} {
-    if {[catch {wm deiconify $w}]} {
-	eval "$type $w $args"
-    } else {
+    if {[winfo exists $w]} {
 	eval "$w config $args"
+	wm deiconify $w
+    } else {
+	eval "$type $w $args"
     }
 }
 
 
-# display the hyperhelp window with the given file
-
-#proc utilHelp {{file Intro.hlp}} {
-#    global app_library
-#    if {![winfo exists .hyperhelp]} {
-#	hyperhelp_init app $app_library/help
-#	wm minsize .hyperhelp 500 500
-#    }
-#    hyperhelp_file $file
-#}
+# temp: until Tk4.0
 
+proc cget {w option} {
+    return [lindex [$w config $option] 4]
+}
